import importlib.util
import os
import pathlib

import numpy as np
import pytest
import wapordl
import wapordl.bounding_boxes
import wapordl.main
import wapordl.unit_convertor
from wapordl import (
    wapor_map,
    wapor_ts,
)

optional_packages = ["dask", "rioxarray", "xarray"]
use_xarray = all(
    [not isinstance(importlib.util.find_spec(x), type(None)) for x in optional_packages]
)
if use_xarray:
    import xarray as xr

module_path = wapordl.__path__[0]
assert "conda" not in module_path

test_data_folder = pathlib.Path(module_path).parent / "test_data"

@pytest.mark.skipif(condition=not use_xarray, reason="XArray not available")
def test_xarray_1(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    # print(tmp_path)
    wapordl.unit_convertor.use_xarray = False
    xx_1 = wapor_map(
        region,
        "L2-AETI-M",
        period,
        os.path.join(tmp_path, "no_xarray"),
        unit_conversion="day",
        overview=3,
    )
    wapordl.unit_convertor.use_xarray = True
    xx_2 = wapor_map(
        region,
        "L2-AETI-M",
        period,
        os.path.join(tmp_path, "ya_xarray"),
        unit_conversion="day",
        overview=3,
    )
    x1 = xr.open_dataset(xx_1)
    x2 = xr.open_dataset(xx_2)
    assert x1.mean() == x2.mean()

@pytest.mark.skipif(condition=not use_xarray, reason="XArray not available")
def test_xarray_2():

    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    wapordl.unit_convertor.use_xarray = False
    dff_1 = wapor_ts(region, "L2-AETI-M", period, unit_conversion="day", overview=3)
    wapordl.unit_convertor.use_xarray = True
    dff_2 = wapor_ts(region, "L2-AETI-M", period, unit_conversion="day", overview=3)
    assert dff_1.equals(dff_2)

@pytest.mark.skipif(condition=not use_xarray, reason="XArray not available")
def test_big_data(tmp_path):
    # BIG DATA, this crashes without Dask.
    wapordl.unit_convertor.use_xarray = True
    wapor_map(
        "ENO",
        "L3-T-D",
        ["2021-12-01", "2021-12-31"],
        os.path.join(tmp_path, "big_xarray"),
        unit_conversion="dekad",
    )

@pytest.mark.skipif(condition=not use_xarray, reason="XArray not available")
def test_summation(tmp_path):
    bb = [30.2, 28.6, 31.3, 30.5]
    period = ["2018-01-01", "2018-12-31"]

    fp_a_nc = wapordl.wapor_map(bb, "L2-AETI-A", period, tmp_path, extension=".nc")
    fp_d_nc = wapordl.wapor_map(bb, "L2-AETI-D", period, tmp_path, extension=".nc")
    fp_dd_nc = wapordl.wapor_map(
        bb, "L2-AETI-D", period, tmp_path, extension=".nc", unit_conversion="dekad"
    )

    ds_d = xr.open_dataset(fp_d_nc, decode_coords="all")
    coords = [
        np.datetime64(da.attrs["start_date"], "ns") for da in ds_d.data_vars.values()
    ]
    da_d = ds_d.to_array("time").assign_coords({"time": coords})
    length = xr.where(da_d["time"].dt.day != 21, 10, da_d["time"].dt.daysinmonth - 20)
    da_d = (da_d * length).sum(dim="time")

    ds_dd = xr.open_dataset(fp_dd_nc, decode_coords="all")
    coords = [
        np.datetime64(da.attrs["start_date"], "ns") for da in ds_dd.data_vars.values()
    ]
    da_dd = ds_dd.to_array("time").assign_coords({"time": coords})
    da_dd = da_dd.sum(dim="time")

    ds_a = xr.open_dataset(fp_a_nc, decode_coords="all")
    da_a = ds_a["Band1"]

    assert abs((da_a - da_d).mean().values) < 0.00001
    assert abs((da_a - da_dd).mean().values) < 0.00001
    assert abs((da_d - da_dd).mean().values) < 0.00001



import os
import pathlib

import pytest

import wapordl
import wapordl.products
import wapordl.products.wapor3
import wapordl.toolbox
import wapordl.toolbox.ogr_gdal
import wapordl.variable_descriptions

module_path = wapordl.__path__[0]
assert "conda" not in module_path

test_data_folder = pathlib.Path(module_path).parent / "test_data"


def test_extension_keyword():
    with pytest.raises(ValueError) as e:
        urls = []
        out_fn = "blabla.weirdext"
        _ = wapordl.toolbox.ogr_gdal.cog_dl(urls, out_fn)
    assert "Please use one of " in str(e.value)


def test_2():
    with pytest.raises(ValueError) as e:
        _ = wapordl.variable_descriptions.date_func(
            "https://storage.googleapis.com/fao-gismgr-wapor-3-data/DATA/WAPOR-3/MOSAICSET/L3-T-D/WAPOR-3.L3-T-D.BKA.2021-01-D1.tif",
            "W",
        )
    assert "Invalid temporal resolution." in str(e.value)


def test_3():
    with pytest.raises(ValueError) as e:
        _ = wapordl.variable_descriptions.collect_metadata("L4-AETI-D")
    assert "Invalid variable name" in str(e.value)


def test_4():
    with pytest.raises(ValueError) as e:
        _ = wapordl.products.wapor3.generate_urls(
            "L4-AETI-D", l3_region=None, period=None
        )
    assert "Invalid level " in str(e.value)


def test_5():
    period = ["2021-01-12", "2021-01-25"]
    overview = 3
    region = os.path.join(test_data_folder, "1237500.geojson")
    with pytest.raises(ValueError) as e:
        _ = wapordl.wapor_ts(
            region.replace(".geojson", "blabla.geojson"), "L1-AETI-A", period, overview
        )
    assert "Geojson file not found." in str(e.value)


def test_6():
    period = ["2021-01-12", "2021-01-25"]
    overview = 3
    with pytest.raises(ValueError) as e:
        _ = wapordl.wapor_ts([25, -17, 24, -16], "L1-AETI-A", period, overview)
    assert "Invalid bounding box." in str(e.value)


def test_7():
    period = ["2021-01-12", "2021-01-25"]
    overview = 3
    with pytest.raises(ValueError) as e:
        _ = wapordl.wapor_ts((500.0, "lala"), "L3-AETI-A", period, overview)
    assert "Invalid value for" in str(e.value)


def test_8():
    overview = 3
    region = os.path.join(test_data_folder, "1237500.geojson")
    with pytest.raises(ValueError) as e:
        _ = wapordl.wapor_ts(
            region, "L1-AETI-A", ["2021-01-15", "2021-01-01"], overview
        )
    assert "Invalid period." in str(e.value)


def test_11():
    overview = 3
    nodata_period = ["2015-01-01", "2016-01-01"]
    region = os.path.join(test_data_folder, "1237500.geojson")
    with pytest.raises(ValueError) as e:
        _ = wapordl.wapor_ts(region, "L2-AETI-D", nodata_period, overview)
    assert "No files found for selected region, variable and period." in str(e.value)


def test_12(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region = os.path.join(test_data_folder, "1237500.geojson")
    with pytest.raises(ValueError) as e:
        _ = wapordl.wapor_map(
            region, "L2-AETI-D", period, tmp_path, unit_conversion="pentad"
        )
    assert "Please select one of " in str(e.value)


def test_13():
    period = ["2021-01-12", "2021-01-25"]
    overview = 3
    region = os.path.join(test_data_folder, "1237500.geojson")
    with pytest.raises(ValueError) as e:
        _ = wapordl.wapor_ts(
            region, "L2-AETI-D", period, overview, unit_conversion="pentad"
        )
    assert "Please select one of " in str(e.value)


def test_14(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    with pytest.raises(ValueError) as e:
        bb_south_america = [-68.203125, -18.979026, -55.371094, -9.839170]
        _ = wapordl.wapor_map(bb_south_america, "L2-T-D", period, tmp_path)
    assert "has no overlap with the datasets" in str(e.value)


def test_15(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    with pytest.raises(ValueError) as e:
        region_FAIL = os.path.join(test_data_folder, "test_FAIL.geojson")
        _ = wapordl.wapor_map(region_FAIL, "L3-T-D", period, tmp_path)
    assert "`region` can't be linked to any L3 region." in str(e.value)



import os

import pandas as pd

from wapordl.toolbox.query import collect_responses


def collect_metadata(variable: str) -> dict:
    """Queries `long_name`, `units` and `source` for a given WaPOR variable code.

    Parameters
    ----------
    variable : str
        Name of variable, e.g. `L3-AETI-D`.

    Returns
    -------
    dict
        Metadata for the variable.

    Raises
    ------
    ValueError
        No valid variable name given.
    """

    if variable in AGERA5_VARS.keys():
        return AGERA5_VARS[variable]

    if variable in WAPOR3_VARS.keys():
        return WAPOR3_VARS[variable]

    if "L1" in variable:
        base_url = (
            "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/mapsets"
        )
    elif "L2" in variable:
        base_url = (
            "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/mapsets"
        )
    elif "L3" in variable:
        base_url = "https://data.apps.fao.org/gismgr/api/v2/catalog/workspaces/WAPOR-3/mosaicsets"
    else:
        raise ValueError(f"Invalid variable name {variable}.")  # NOTE: TESTED
    info = ["code", "measureCaption", "measureUnit"]
    var_codes = {
        x[0]: {"long_name": x[1], "units": x[2]}
        for x in collect_responses(base_url, info=info)
    }

    return var_codes[variable]


def date_func(url: str, tres: str) -> dict:
    """Determines start and end dates from a string a given temporal resolution, as well
    as the number of days between the two dates.

    Parameters
    ----------
    url : str
        URL linking to a resource.
    tres : str
        One of `"E"` (daily), `"D"` (dekadal), `"M"` (monthly), `"A"` (annual).

    Returns
    -------
    dict
        Dates and related information for a resource URL.

    Raises
    ------
    ValueError
        No valid `tres` given.
    """
    if tres == "D":
        year, month, dekad = os.path.split(url)[-1].split(".")[-2].split("-")
        start_day = {
            "D1": "01",
            "D2": "11",
            "D3": "21",
            "1": "01",
            "2": "11",
            "3": "21",
        }[dekad]
        start_date = f"{year}-{month}-{start_day}"
        end_day = {
            "D1": "10",
            "D2": "20",
            "D3": pd.Timestamp(start_date).daysinmonth,
            "1": "10",
            "2": "20",
            "3": pd.Timestamp(start_date).daysinmonth,
        }[dekad]
        end_date = f"{year}-{month}-{end_day}"
    elif tres == "M":
        year, month = os.path.split(url)[-1].split(".")[-2].split("-")
        start_date = f"{year}-{month}-01"
        end_date = f"{year}-{month}-{pd.Timestamp(start_date).days_in_month}"
    elif tres == "A":
        year = os.path.split(url)[-1].split(".")[-2]
        start_date = f"{year}-01-01"
        end_date = f"{year}-12-31"
    elif tres == "E":
        year, month, start_day = os.path.split(url)[-1].split(".")[-2].split("-")
        start_date = end_date = f"{year}-{month}-{start_day}"
    else:
        raise ValueError("Invalid temporal resolution.")  # NOTE: TESTED

    number_of_days = (
        pd.Timestamp(end_date) - pd.Timestamp(start_date) + pd.Timedelta(1, "D")
    ).days

    date_md = {
        "start_date": start_date,
        "end_date": end_date,
        "number_of_days": number_of_days,
        "temporal_resolution": {"E": "Day", "D": "Dekad", "M": "Month", "A": "Year"}[
            tres
        ],
    }

    if tres == "E":
        dekad = min(3, ((int(start_day) - 1) // 10) + 1)
        days_in_dekad = {1: 10, 2: 10, 3: pd.Timestamp(start_date).daysinmonth - 20}[
            dekad
        ]
        date_md["days_in_dekad"] = days_in_dekad

    return date_md


AGERA5_VARS = {
    "AGERA5-ET0-E": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/day",
        "source": "FAO56 with agERA5",
    },
    "AGERA5-ET0-D": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/dekad",
        "source": "FAO56 with agERA5",
    },
    "AGERA5-ET0-M": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/month",
        "source": "FAO56 with agERA5",
    },
    "AGERA5-ET0-A": {
        "long_name": "Reference Evapotranspiration",
        "units": "mm/year",
        "source": "FAO56 with agERA5",
    },
    "AGERA5-TMIN-E": {
        "long_name": "Minimum Air Temperature (2m)",
        "units": "K",
        "source": "agERA5",
    },
    "AGERA5-TMAX-E": {
        "long_name": "Maximum Air Temperature (2m)",
        "units": "K",
        "source": "agERA5",
    },
    "AGERA5-SRF-E": {
        "long_name": "Solar Radiation",
        "units": "J/m2/day",
        "source": "agERA5",
    },
    "AGERA5-WS-E": {"long_name": "Wind Speed", "units": "m/s", "source": "agERA5"},
    "AGERA5-RH06-E": {
        "long_name": "Relative humidity at 06h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
    },
    "AGERA5-RH09-E": {
        "long_name": "Relative humidity at 09h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
    },
    "AGERA5-RH12-E": {
        "long_name": "Relative humidity at 12h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
    },
    "AGERA5-RH15-E": {
        "long_name": "Relative humidity at 15h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
    },
    "AGERA5-RH18-E": {
        "long_name": "Relative humidity at 18h (local time, 2m)",
        "units": "%",
        "source": "agERA5",
    },
    "AGERA5-PF-E": {
        "long_name": "Precipitation",
        "units": "mm/day",
        "source": "agERA5",
    },
    "AGERA5-PF-D": {
        "long_name": "Precipitation",
        "units": "mm/dekad",
        "source": "agERA5",
    },
    "AGERA5-PF-M": {
        "long_name": "Precipitation",
        "units": "mm/month",
        "source": "agERA5",
    },
    "AGERA5-PF-A": {
        "long_name": "Precipitation",
        "units": "mm/year",
        "source": "agERA5",
    },
}

WAPOR3_VARS = {
    "L1-AETI-A": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/year",
    },
    "L1-AETI-D": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/day",
    },
    "L1-AETI-M": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/month",
    },
    "L1-E-A": {"long_name": "Evaporation", "units": "mm/year"},
    "L1-E-D": {"long_name": "Evaporation", "units": "mm/day"},
    "L1-GBWP-A": {"long_name": "Gross Biomass Water Productivity", "units": "kg/m³"},
    "L1-I-A": {"long_name": "Interception", "units": "mm/year"},
    "L1-I-D": {"long_name": "Interception", "units": "mm/day"},
    "L1-NBWP-A": {"long_name": "Net Biomass Water Productivity", "units": "kg/m³"},
    "L1-NPP-D": {"long_name": "Net Primary Production", "units": "gC/m²/day"},
    "L1-NPP-M": {"long_name": "Net Primary Production", "units": "gC/m²/month"},
    "L1-PCP-A": {"long_name": "Precipitation", "units": "mm/year"},
    "L1-PCP-D": {"long_name": "Precipitation", "units": "mm/day"},
    "L1-PCP-E": {"long_name": "Precipitation", "units": "mm/day"},
    "L1-PCP-M": {"long_name": "Precipitation", "units": "mm/month"},
    "L1-QUAL-LST-D": {"long_name": "Quality Land Surface Temperature", "units": "d"},
    "L1-QUAL-NDVI-D": {
        "long_name": "Quality of Normalized Difference Vegetation Index",
        "units": "d",
    },
    "L1-RET-A": {"long_name": "Reference Evapotranspiration", "units": "mm/year"},
    "L1-RET-D": {"long_name": "Reference Evapotranspiration", "units": "mm/day"},
    "L1-RET-E": {"long_name": "Reference Evapotranspiration", "units": "mm/day"},
    "L1-RET-M": {"long_name": "Reference Evapotranspiration", "units": "mm/month"},
    "L1-RSM-D": {"long_name": "Relative Soil Moisture", "units": "%"},
    "L1-T-A": {"long_name": "Transpiration", "units": "mm/year"},
    "L1-T-D": {"long_name": "Transpiration", "units": "mm/day"},
    "L1-TBP-A": {"long_name": "Total Biomass Production", "units": "kg/ha"},
    "L2-AETI-A": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/year",
    },
    "L2-AETI-D": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/day",
    },
    "L2-AETI-M": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/month",
    },
    "L2-E-A": {"long_name": "Evaporation", "units": "mm/year"},
    "L2-E-D": {"long_name": "Evaporation", "units": "mm/day"},
    "L2-GBWP-A": {"long_name": "Gross Biomass Water Productivity", "units": "kg/m³"},
    "L2-I-A": {"long_name": "Interception", "units": "mm/year"},
    "L2-I-D": {"long_name": "Interception", "units": "mm/day"},
    "L2-NBWP-A": {"long_name": "Net Biomass Water Productivity", "units": "kg/m³"},
    "L2-NPP-D": {"long_name": "Net Primary Production", "units": "gC/m²/day"},
    "L2-NPP-M": {"long_name": "Net Primary Production", "units": "gC/m²/month"},
    "L2-QUAL-NDVI-D": {
        "long_name": "Quality of Normalized Difference Vegetation Index",
        "units": "d",
    },
    "L2-RSM-D": {"long_name": "Relative Soil Moisture", "units": "%"},
    "L2-T-A": {"long_name": "Transpiration", "units": "mm/year"},
    "L2-T-D": {"long_name": "Transpiration", "units": "mm/day"},
    "L2-TBP-A": {"long_name": "Total Biomass Production", "units": "kg/ha"},
    "L3-AETI-A": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/year",
    },
    "L3-AETI-D": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/day",
    },
    "L3-AETI-M": {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/month",
    },
    "L3-E-A": {"long_name": "Evaporation", "units": "mm/year"},
    "L3-E-D": {"long_name": "Evaporation", "units": "mm/day"},
    "L3-GBWP-A": {"long_name": "Gross Biomass Water Productivity", "units": "kg/m³"},
    "L3-I-A": {"long_name": "Interception", "units": "mm/year"},
    "L3-I-D": {"long_name": "Interception", "units": "mm/day"},
    "L3-NBWP-A": {"long_name": "Net Biomass Water Productivity", "units": " kg/m³"},
    "L3-NPP-D": {"long_name": "Net Primary Production", "units": "gC/m²/day"},
    "L3-NPP-M": {"long_name": "Net Primary Production", "units": "gC/m²/month"},
    "L3-QUAL-NDVI-D": {
        "long_name": "Quality of Normalized Difference Vegetation Index",
        "units": "d",
    },
    "L3-RSM-D": {"long_name": "Relative Soil Moisture", "units": "%"},
    "L3-T-A": {"long_name": "Transpiration", "units": "mm/year"},
    "L3-T-D": {"long_name": "Transpiration", "units": "mm/day"},
    "L3-TBP-A": {"long_name": "Total Biomass Production", "units": "kg/ha"},
}



import os
from typing import List, Tuple

import numpy as np
import pandas as pd

import wapordl.toolbox.query as query
from wapordl.variable_descriptions import AGERA5_VARS


def generate_urls(
    variable: str, period: List[str] | None = None, check_urls: bool = True
) -> Tuple[str]:
    """Find resource URLs for an agERA5 variable for a specified period.

    Parameters
    ----------
    variable : str
        Name of the variable.
    period : list, optional
        Start and end date in between which resource URLs will be searched, by default None.
    check_urls : bool, optional
        Perform additional checks to test if the found URLs are valid, by default True.

    Returns
    -------
    tuple
        Resource URLs.

    Raises
    ------
    ValueError
        Invalid variable selected.
    ValueError
        Invalid temporal resolution.

    Notes
    -----
    https://data.apps.fao.org/static/data/index.html?prefix=static%2Fdata%2Fc3s%2FMAPSET%2FAGERA5-ET0-D
    """
    level, var_code, tres = variable.split("-")

    if variable not in AGERA5_VARS.keys():
        raise ValueError(
            f"Invalid variable `{variable}`, choose one from `{AGERA5_VARS.keys()}`."
        )

    max_date = pd.Timestamp.now() - pd.Timedelta(days=25)
    if isinstance(period, type(None)):
        period = ["1979-01-01", max_date.strftime("%Y-%m-%d")]

    base_url = "https://data.apps.fao.org/static/data/c3s/MAPSET"
    urls = list()
    if tres == "E":
        x_filtered = make_daily_dates(period, max_date=max_date)
        for x in x_filtered:
            url = os.path.join(
                base_url,
                f"{level}-{var_code}",
                f"C3S.{level}-{var_code}.{x.strftime('%Y-%m-%d')}.tif",
            )
            urls.append(url)
    elif tres == "D":
        x_filtered = make_dekad_dates(period, max_date=max_date)
        for x in x_filtered:
            dekad = {1: 1, 11: 2, 21: 3}[x.day]
            url = os.path.join(
                base_url,
                variable,
                f"C3S.{variable}.{x.year}-{x.month:>02}-D{dekad}.tif",
            )
            urls.append(url)
    elif tres == "M":
        x_filtered = make_monthly_dates(period, max_date=max_date)
        for x in x_filtered:
            url = os.path.join(
                base_url, variable, f"C3S.{variable}.{x.year}-{x.month:>02}.tif"
            )
            urls.append(url)
    elif tres == "A":
        x_filtered = make_annual_dates(period, max_date=max_date)
        for x in x_filtered:
            url = os.path.join(base_url, variable, f"C3S.{variable}.{x.year}.tif")
            urls.append(url)
    else:
        raise ValueError(f"Invalid temporal resolution `{tres}`.")

    if check_urls:
        urls = query.check_urls(urls)

    return tuple(sorted(urls))


def make_dekad_dates(
    period: List[str], max_date: pd.Timestamp | None = None
) -> List[pd.Timestamp]:
    """Make a list of dekadal timestamps between a start and end date.

    Parameters
    ----------
    period : list
        Start and end date in between which the dekadal timestamps will be generated.
    max_date : pd.Timestamp, optional
        Choose the earliest date between the end of `period` and `max_date`, by default None.

    Returns
    -------
    list
        Dekadal timestamps between the given start and end date.
    """
    period_ = [pd.Timestamp(x) for x in period]
    if isinstance(max_date, pd.Timestamp):
        period_[1] = min(period_[1], max_date)
    syear = period_[0].year
    smonth = period_[0].month
    eyear = period_[1].year
    emonth = period_[1].month
    x1 = pd.date_range(f"{syear}-{smonth}-01", f"{eyear}-{emonth}-01", freq="MS")
    x2 = x1 + pd.Timedelta("10 days")
    x3 = x1 + pd.Timedelta("20 days")
    x = np.sort(np.concatenate((x1, x2, x3)))
    x_filtered = [pd.Timestamp(x_) for x_ in x if x_ >= period_[0] and x_ < period_[1]]
    return x_filtered


def make_monthly_dates(
    period: List[str], max_date: pd.Timestamp | None = None
) -> List[pd.Timestamp]:
    """Make a list of monthly timestamps between a start and end date.

    Parameters
    ----------
    period : list
        Start and end date in between which the monthly timestamps will be generated.
    max_date : pd.Timestamp, optional
        Choose the earliest date between the end of `period` and `max_date`, by default None.

    Returns
    -------
    list
        Monthly timestamps between the given start and end date.
    """
    period_ = [pd.Timestamp(x) for x in period]
    period_[0] = pd.Timestamp(f"{period_[0].year}-{period_[0].month}-01")
    if isinstance(max_date, pd.Timestamp):
        period_[1] = min(period_[1], max_date)
    x1 = pd.date_range(period_[0], period_[1], freq="MS")
    x_filtered = [pd.Timestamp(x_) for x_ in x1]
    return x_filtered


def make_annual_dates(
    period: List[str], max_date: pd.Timestamp | None = None
) -> List[pd.Timestamp]:
    """Make a list of annual timestamps between a start and end date.

    Parameters
    ----------
    period : list
        Start and end date in between which the annual timestamps will be generated.
    max_date : pd.Timestamp, optional
        Choose the earliest date between the end of `period` and `max_date`, by default None.

    Returns
    -------
    list
        Annual timestamps between the given start and end date.
    """
    period_ = [pd.Timestamp(x) for x in period]
    period_[0] = pd.Timestamp(f"{period_[0].year}-01-01")
    if isinstance(max_date, pd.Timestamp):
        period_[1] = min(period_[1], max_date)
    try:
        x1 = pd.date_range(period_[0], period_[1], freq="YE-JAN")
    except ValueError as e:
        if "Invalid frequency: YE-JAN" in str(e):
            x1 = pd.date_range(period_[0], period_[1], freq="A-JAN")
        else: 
            raise(e)
    x_filtered = [pd.Timestamp(x_) for x_ in x1]
    return x_filtered


def make_daily_dates(
    period: List[str], max_date: pd.Timestamp | None = None
) -> List[pd.Timestamp]:
    """Make a list of daily timestamps between a start and end date.

    Parameters
    ----------
    period : list
        Start and end date in between which the daily timestamps will be generated.
    max_date : pd.Timestamp, optional
        Choose the earliest date between the end of `period` and `max_date`, by default None.

    Returns
    -------
    list
        Daily timestamps between the given start and end date.
    """
    period_ = [pd.Timestamp(x) for x in period]
    if isinstance(max_date, pd.Timestamp):
        period_[1] = min(period_[1], max_date)
    x1 = pd.date_range(period_[0], period_[1], freq="D")
    x_filtered = [pd.Timestamp(x_) for x_ in x1]
    return x_filtered



import os
import pathlib

import numpy as np
import pytest

import wapordl
import wapordl.products
import wapordl.products.wapor3
import wapordl.toolbox
import wapordl.toolbox.ogr_gdal
import wapordl.variable_descriptions

module_path = wapordl.__path__[0]
assert "conda" not in module_path

test_data_folder = pathlib.Path(module_path).parent / "test_data"

vectors_ids = [
    (os.path.join(test_data_folder, "polygon_EPSG32636.gpkg"), "name"),
    (os.path.join(test_data_folder, "polygon_EPSG32636.gpkg"), "bla"),
    (os.path.join(test_data_folder, "multipolygon_EPSG4326.gpkg"), "name"),
    (os.path.join(test_data_folder, "multipolygon_EPSG4326.gpkg"), "bla"),
]


def test_main1():
    variable = "L3-T-D"
    period = ["2021-01-01", "2021-01-31"]
    overview = "NONE"
    region = os.path.join(test_data_folder, "polygon_EPSG32636.gpkg")
    identifier = "name"
    df1 = wapordl.wapor_ts(
        region, variable, period, identifier=identifier, overview=overview
    )
    assert df1.columns.size == 7
    assert identifier in list(df1.columns)
    assert np.isfinite(df1["mean"].dropna().mean())
    assert overview in df1.attrs["overview"]
    df2 = wapordl.wapor_ts(
        region, variable, period, identifier=identifier, overview=overview, n_threads=5
    )
    assert df1.equals(df2)


def test_main2():
    variable = "L3-T-D"
    period = ["2021-01-01", "2021-01-31"]
    overview = "NONE"
    region = os.path.join(test_data_folder, "polygon_EPSG32636.gpkg")
    identifier = "name"
    overview = "AUTO"
    df2 = wapordl.wapor_ts(
        region, variable, period, identifier=identifier, overview=overview
    )
    assert df2.columns.size == 7
    assert identifier in list(df2.columns)
    assert np.isfinite(df2["mean"].dropna().mean())
    assert overview in df2.attrs["overview"]
    assert not any(df2["start_date"].isnull())
    df3 = wapordl.wapor_ts(
        region, variable, period, identifier=identifier, overview=overview, n_threads=5
    )
    assert df2.equals(df3)


def test_main3():
    variable = "L3-T-D"
    period = ["2021-01-01", "2021-01-31"]
    overview = "NONE"
    region = os.path.join(test_data_folder, "polygon_EPSG32636.gpkg")
    identifier = None
    overview = "NONE"
    df3 = wapordl.wapor_ts(
        region, variable, period, identifier=identifier, overview=overview
    )
    assert df3.columns.size == 6
    assert overview in df3.attrs["overview"]
    df4 = wapordl.wapor_ts(
        region, variable, period, identifier=identifier, overview=overview, n_threads=5
    )
    assert df4.equals(df3)


def test_main4():
    variable = "L3-T-D"
    period = ["2021-01-01", "2021-01-31"]
    overview = "NONE"
    region = os.path.join(test_data_folder, "polygon_EPSG32636.gpkg")
    identifier = "name"
    region = "GEZ"
    overview = "AUTO"
    df4 = wapordl.wapor_ts(
        region, variable, period, identifier=identifier, overview=overview
    )
    assert df4.columns.size == 6


@pytest.mark.parametrize("vector,identifier", vectors_ids)
def test_base(vector, identifier):
    raster = os.path.join(test_data_folder, "GEZ_L3-AETI-D_1_none.tif")
    out = wapordl.toolbox.ogr_gdal.zonal_stats(raster, vector, identifier)
    assert out.size > 0
    assert out["mean"].dropna().size > 0
    assert np.isfinite(out["mean"].dropna().mean())


def test_invalid_id():
    raster = os.path.join(test_data_folder, "GEZ_L3-AETI-D_1_none.tif")
    vector = os.path.join(test_data_folder, "polygon_EPSG32636.gpkg")
    identifier = "XX"
    with pytest.raises(ValueError) as e:
        _ = wapordl.toolbox.ogr_gdal.zonal_stats(raster, vector, identifier)
    assert "not found in field names" in str(e.value)


def test_invalid_geomtype():
    raster = os.path.join(test_data_folder, "GEZ_L3-AETI-D_1_none.tif")
    vector = os.path.join(test_data_folder, "point_EPSG4326.gpkg")
    identifier = "name"
    with pytest.raises(NotImplementedError) as e:
        _ = wapordl.toolbox.ogr_gdal.zonal_stats(raster, vector, identifier)
    assert "`zonal_stats` for geometry type" in str(e.value)


if __name__ == "__main__":
    vector = vectors_ids[0][0]
    identifier = vectors_ids[0][1]



import logging
from typing import List

import requests


def check_urls(urls: List[str]) -> List[str]:
    """Check if URLS exist and filter out non-valid ones.

    Parameters
    ----------
    urls : List[str]
        ULRS to check.

    Returns
    -------
    List[str]
        URLS with non-valid ones filtered out.
    """
    for url in urls.copy():
        try:
            x = requests.get(url, stream=True)
            x.raise_for_status()
        except requests.exceptions.HTTPError:
            logging.debug(f"Invalid url detected, removing `{url}`.")
            urls.remove(url)
    return urls


def collect_responses(url: str, info: List[str] = ["code"]) -> list:
    """Calls GISMGR2.0 API and collects responses.

    Parameters
    ----------
    url : str
        URL to get.
    info : list, optional
        Used to filter the response, set to `None` to keep everything, by default `["code"]`.

    Returns
    -------
    list
        The responses.
    """
    data = {"links": [{"rel": "next", "href": url}]}
    output = list()
    while "next" in [x["rel"] for x in data["links"]]:
        url_ = [x["href"] for x in data["links"] if x["rel"] == "next"][0]
        response = requests.get(url_)
        response.raise_for_status()
        data = response.json()["response"]
        if isinstance(info, list) and "items" in data.keys():
            output += [tuple(x.get(y) for y in info) for x in data["items"]]
        elif "items" in data.keys():
            output += data["items"]
        else:
            output.append(data)
    if isinstance(info, list):
        try:
            output = sorted(output)
        except TypeError:
            output = output
    return output



import glob
import os

import pathlib

import numpy as np
import pandas as pd
import pytest
from osgeo import gdal, osr

import wapordl
import wapordl.bounding_boxes
import wapordl.main
import wapordl.overview_selector
import wapordl.unit_convertor
from wapordl import (
    wapor_map,
    wapor_ts,
)

module_path = wapordl.__path__[0]
assert "conda" not in module_path

test_data_folder = pathlib.Path(module_path).parent / "test_data"


#####
# PART 2
#####

fhs = [
    "detector_shapes/box_1.geojson",
    "detector_shapes/box_3.geojson",
    "detector_shapes/box_5.geojson",
    "detector_shapes/box_NONE.geojson",
]


@pytest.mark.parametrize("fh", fhs)
def test_overview_detector_1(fh, tmp_path):
    fh = os.path.join(test_data_folder, fh)
    x1 = wapordl.wapor_map(
        fh, "L1-T-D", ["2021-01-01", "2021-01-01"], tmp_path, overview="auto"
    )
    assert str(os.path.split(x1)[-1].split("_")[-2]) == os.path.split(fh)[-1].split(
        "_"
    )[-1].replace(".geojson", "")


def test_overview_detector_2(tmp_path):
    fh = os.path.join(test_data_folder, "detector_shapes/star.geojson")
    pngs = glob.glob(os.path.join(tmp_path, "*.png"))
    for png in pngs:
        os.remove(png)
    assert len(glob.glob(os.path.join(tmp_path, "*.png"))) == 0
    x5 = wapordl.wapor_map(
        fh,
        "L1-T-D",
        ["2021-01-01", "2021-01-01"],
        tmp_path,
        overview="AUTO",
        make_plots=tmp_path,
    )
    if wapordl.overview_selector.use_plt:
        assert len(glob.glob(os.path.join(tmp_path, "*.png"))) > 0
    pngs = glob.glob(os.path.join(tmp_path, "*.png"))
    for png in pngs:
        os.remove(png)

    fh = os.path.join(test_data_folder, "detector_shapes/tiny_box.geojson")
    x6 = wapordl.wapor_map(
        fh,
        "L1-T-D",
        ["2021-01-01", "2021-01-01"],
        os.path.join(tmp_path, "big_xarray"),
        overview="AUTO",
        max_error=0.001,
    )
    x6_overview = os.path.split(x6)[-1].split("_")[-2]
    if x6_overview == "NONE":
        x6_overview = -1
    assert int(os.path.split(x5)[-1].split("_")[-2]) > x6_overview


def test_overview_detector_3(tmp_path):
    fh = os.path.join(test_data_folder, "test_MUV.geojson")
    pngs = glob.glob(os.path.join(tmp_path, "*.png"))
    for png in pngs:
        os.remove(png)
    df1 = wapordl.wapor_ts(
        fh,
        "L3-AETI-D",
        ["2021-01-01", "2021-02-01"],
        overview="auto",
        make_plots=tmp_path,
        max_error=0.3,
    )
    assert df1["mean"].mean() > 2.5
    pngs = glob.glob(os.path.join(tmp_path, "*.png"))
    for png in pngs:
        os.remove(png)


#####
# AGERA5
#####

def test_agera5_region_code(tmp_path):
    variable = "AGERA5-PF-D"
    region = "BKA"
    period = ['2018-01-01', '2018-02-01']
    fh = wapor_map(region, variable, period, tmp_path)
    info =gdal.Info(fh, format="json")
    assert info["size"] == [6, 5]


def test_agera5_1(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period_agera5 = [
        (pd.Timestamp.now() - pd.Timedelta(days=80)).strftime("%Y-%m-%d"),
        (pd.Timestamp.now() - pd.Timedelta(days=40)).strftime("%Y-%m-%d"),
    ]
    _ = wapor_map(
        region,
        "AGERA5-ET0-D",
        period_agera5,
        tmp_path,
        extension=".nc",
        unit_conversion="dekad",
    )


def test_agera5_2(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period_agera5 = [
        (pd.Timestamp.now() - pd.Timedelta(days=80)).strftime("%Y-%m-%d"),
        (pd.Timestamp.now() - pd.Timedelta(days=40)).strftime("%Y-%m-%d"),
    ]
    _ = wapor_map(
        region,
        "AGERA5-ET0-D",
        period_agera5,
        tmp_path,
        extension=".nc",
        unit_conversion="day",
    )


def test_agera5_3(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period_agera5 = [
        (pd.Timestamp.now() - pd.Timedelta(days=80)).strftime("%Y-%m-%d"),
        (pd.Timestamp.now() - pd.Timedelta(days=40)).strftime("%Y-%m-%d"),
    ]
    _ = wapor_map(
        region,
        "AGERA5-ET0-E",
        period_agera5,
        tmp_path,
        extension=".nc",
        unit_conversion="dekad",
    )


def test_agera5_4(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period_agera5 = [
        (pd.Timestamp.now() - pd.Timedelta(days=80)).strftime("%Y-%m-%d"),
        (pd.Timestamp.now() - pd.Timedelta(days=40)).strftime("%Y-%m-%d"),
    ]
    _ = wapor_map(region, "AGERA5-ET0-M", period_agera5, tmp_path)


def test_agera5_5():
    region = os.path.join(test_data_folder, "1237500.geojson")
    period_agera5 = [
        (pd.Timestamp.now() - pd.Timedelta(days=80)).strftime("%Y-%m-%d"),
        (pd.Timestamp.now() - pd.Timedelta(days=40)).strftime("%Y-%m-%d"),
    ]
    _ = wapor_ts(
        region, "AGERA5-TMAX-E", period_agera5, overview=3, unit_conversion="day"
    )


def test_agera5_6():
    region = os.path.join(test_data_folder, "1237500.geojson")
    period_agera5 = [
        (pd.Timestamp.now() - pd.Timedelta(days=80)).strftime("%Y-%m-%d"),
        (pd.Timestamp.now() - pd.Timedelta(days=40)).strftime("%Y-%m-%d"),
    ]
    _ = wapor_ts(
        region, "AGERA5-TMIN-E", period_agera5, overview="NONE", unit_conversion="month"
    )


def test_agera5_7():
    region = os.path.join(test_data_folder, "1237500.geojson")
    period_agera5 = [
        (pd.Timestamp.now() - pd.Timedelta(days=80)).strftime("%Y-%m-%d"),
        (pd.Timestamp.now() - pd.Timedelta(days=40)).strftime("%Y-%m-%d"),
    ]
    _ = wapor_ts(
        region, "AGERA5-RH12-E", period_agera5, overview=1, unit_conversion="year"
    )


def test_agera5_8(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period_agera5 = ["2022-12-18", pd.Timestamp.now().strftime("%Y-%m-%d")]
    _ = wapor_map(region, "AGERA5-ET0-A", period_agera5, tmp_path)


#####
# GENERAL
#####


def test_general_1():
    overview = 3
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]

    df1 = wapor_ts(region, "L2-AETI-D", period, overview=overview)
    assert np.isclose(df1.iloc[0]["mean"], 0.4291)
    assert df1.iloc[0].start_date == pd.Timestamp("2021-01-11 00:00:00")
    assert df1.attrs == {
        "long_name": "Actual EvapoTranspiration and Interception",
        "units": "mm/day",
        "overview": str(overview),
    }


def test_general_2():
    overview = 3
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    df2 = wapor_ts(region, "L2-AETI-M", period, overview=overview)
    assert df2.attrs["units"] == "mm/month"


def test_general_3():
    overview = 3
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    df3 = wapor_ts(region, "L2-AETI-A", period, overview=overview)
    assert df3.attrs["units"] == "mm/year"


variables = ["L1-AETI-D", "L1-AETI-M", "L1-AETI-A"]


@pytest.mark.parametrize("variable", variables)
def test_general_4(variable):
    overview = 3
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    _ = wapor_ts(region, variable, period, overview=overview)


variables = [
    "L2-AETI-D",
    "L2-AETI-M",
    "L2-AETI-A",
    "L1-AETI-D",
    "L1-AETI-M",
    "L1-AETI-A",
]


@pytest.mark.parametrize("variable", variables)
def test_general_5(variable):
    overview = 3
    bb = [25, -17, 26, -16]
    period = ["2021-01-12", "2021-01-25"]
    _ = wapor_ts(bb, variable, period, overview)


def test_general_6(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    fp1 = wapor_map(region, "L2-AETI-D", period, tmp_path)
    ds = gdal.Open(fp1)
    assert ds.RasterCount == 2
    band = ds.GetRasterBand(1)
    ndv = band.GetNoDataValue()
    scale = band.GetScale()
    array = band.ReadAsArray() * scale
    array[array == ndv * scale] = np.nan
    mean = np.nanmean(array)
    md = band.GetMetadata()
    assert md == {
        "end_date": "2021-01-20",
        "long_name": "Actual EvapoTranspiration and Interception",
        "number_of_days": "10",
        "overview": "NONE",
        "start_date": "2021-01-11",
        "temporal_resolution": "Dekad",
        "units": "mm/day",
    }
    assert mean > 0.0
    assert mean < 15.0
    proj = osr.SpatialReference(wkt=ds.GetProjection())
    assert proj.GetAttrValue("AUTHORITY", 1) == "4326"
    ds = ds.FlushCache()

    fps1a = wapor_map(region, "L2-AETI-D", period, tmp_path, separate_unscale=True)
    ds = gdal.Open(fps1a[1])
    assert ds.RasterCount == 1
    band = ds.GetRasterBand(1)
    ndv = band.GetNoDataValue()
    scale = band.GetScale()
    assert scale == 1 or isinstance(scale, type(None))
    array_X = band.ReadAsArray()
    array_X[array_X == ndv] = np.nan
    mean_X = np.nanmean(array_X)

    ds_ = gdal.Open(fps1a[0])
    band_ = ds_.GetRasterBand(1)
    ndv_ = band_.GetNoDataValue()
    scale_ = band_.GetScale()
    assert scale_ == 1 or isinstance(scale_, type(None))
    array_ = band_.ReadAsArray()
    array_[array_ == ndv_] = np.nan
    mean_ = np.nanmean(array_)
    assert mean_ == mean  # NOTE mean comes from fp1 (without unscaling)
    md = band_.GetMetadata()
    assert md == {
        "end_date": "2021-01-20",
        "long_name": "Actual EvapoTranspiration and Interception",
        "number_of_days": "10",
        "overview": "NONE",
        "start_date": "2021-01-11",
        "temporal_resolution": "Dekad",
        "units": "mm/day",
    }
    assert mean_ > 0.0
    assert mean_ < 15.0
    proj = osr.SpatialReference(wkt=ds.GetProjection())
    assert proj.GetAttrValue("AUTHORITY", 1) == "4326"
    ds = ds.FlushCache()

    fps1b = wapor_map(
        region,
        "L2-AETI-D",
        period,
        tmp_path,
        unit_conversion="dekad",
        separate_unscale=True,
    )
    ds = gdal.Open(fps1b[1])
    assert ds.RasterCount == 1
    band = ds.GetRasterBand(1)
    ndv = band.GetNoDataValue()
    scale = band.GetScale()
    assert scale == 1 or isinstance(scale, type(None))
    array_Y = band.ReadAsArray()
    array_Y[array_Y == ndv] = np.nan
    mean_Y = np.nanmean(array_Y)
    md_Y = band.GetMetadata()
    assert mean_Y / int(md_Y["number_of_days"]) - mean_X < 1e-10


def test_general_7(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    fp2 = wapor_map(region, "L2-AETI-M", period, tmp_path)
    ds = gdal.Open(fp2)
    band = ds.GetRasterBand(1)
    md = band.GetMetadata()
    assert md["units"] == "mm/month"
    ds = ds.FlushCache()


def test_general_8(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    fp3 = wapor_map(region, "L2-AETI-A", period, tmp_path)
    ds = gdal.Open(fp3)
    band = ds.GetRasterBand(1)
    md = band.GetMetadata()
    assert md["units"] == "mm/year"
    ds = ds.FlushCache()


variables = [
    "L1-AETI-D",
    "L1-AETI-M",
    "L1-AETI-A",
]


@pytest.mark.parametrize("variable", variables)
def test_general_9(variable, tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    _ = wapor_map(region, variable, period, tmp_path)


def test_general_10(tmp_path):
    bb = [25, -17, 26, -16]
    period = ["2021-01-12", "2021-01-25"]
    fp7 = wapor_map(bb, "L2-AETI-D", period, tmp_path)
    ds = gdal.Open(fp7)
    geot = ds.GetGeoTransform()
    nx = ds.RasterXSize
    ny = ds.RasterYSize
    assert bb[0] == geot[0]
    assert bb[2] == geot[0] + nx * geot[1]
    assert bb[1] == geot[3] + ny * geot[5]
    assert bb[3] == geot[3]
    band = ds.GetRasterBand(1)
    scale = band.GetScale()
    array = band.ReadAsArray() * scale
    ndv = band.GetNoDataValue()
    array[array == ndv * scale] = np.nan
    mean = np.nanmean(array)
    assert mean > 0.0
    assert mean < 15.0
    proj = osr.SpatialReference(wkt=ds.GetProjection())
    assert proj.GetAttrValue("AUTHORITY", 1) == "4326"
    ds = ds.FlushCache()


variables = [
    "L2-AETI-M",
    "L2-AETI-A",
    "L1-AETI-D",
    "L1-AETI-M",
    "L1-AETI-A",
]


@pytest.mark.parametrize("variable", variables)
def test_general_11(variable, tmp_path):
    bb = [25, -17, 26, -16]
    period = ["2021-01-12", "2021-01-25"]
    _ = wapor_map(bb, variable, period, tmp_path)


def test_general_12(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]
    fp13 = wapor_map(region, "L1-AETI-D", period, tmp_path, extension=".nc")
    info = gdal.Info(fp13, format="json")
    assert len(info["metadata"]["SUBDATASETS"]) == 4
    ds = gdal.Open(info["metadata"]["SUBDATASETS"]["SUBDATASET_1_NAME"])
    band = ds.GetRasterBand(1)
    scale = band.GetScale()
    array = band.ReadAsArray() * scale
    ndv = band.GetNoDataValue()
    array[array == ndv * scale] = np.nan
    mean = np.nanmean(array)
    assert mean > 0.0
    assert mean < 15.0
    proj = osr.SpatialReference(wkt=ds.GetProjection())
    assert proj.GetAttrValue("AUTHORITY", 1) == "4326"
    ds = ds.FlushCache()


def test_general_13(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    fp14 = wapor_map("BKA", "L3-T-D", period, tmp_path)
    ds = gdal.Open(fp14)
    proj = osr.SpatialReference(wkt=ds.GetProjection())
    assert proj.GetAttrValue("AUTHORITY", 1) == "32636"  # NOTE


def test_general_14(tmp_path):
    bbBKA = [35.75, 33.70, 35.82, 33.75]
    period = ["2021-01-12", "2021-01-25"]
    fp15 = wapor_map(bbBKA, "L3-T-D", period, tmp_path)
    assert "bb.BKA" in fp15
    ds = gdal.Open(fp15)
    band = ds.GetRasterBand(1)
    scale = band.GetScale()
    array = band.ReadAsArray() * scale
    ndv = band.GetNoDataValue()
    array[array == ndv * scale] = np.nan
    mean = np.nanmean(array)
    assert mean > 0.0
    assert mean < 15.0
    proj = osr.SpatialReference(wkt=ds.GetProjection())
    assert proj.GetAttrValue("AUTHORITY", 1) == "32636"
    ds = ds.FlushCache()


def test_general_15(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_bekaa = os.path.join(test_data_folder, "test_bekaa.geojson")
    _ = wapor_map(region_bekaa, "L1-T-D", period, tmp_path)


def test_general_16(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_GEZ = os.path.join(test_data_folder, "test_GEZ.geojson")
    _ = wapor_map(region_GEZ, "L3-T-D", period, tmp_path)


def test_general_17(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_GEZ = os.path.join(test_data_folder, "test_GEZ.geojson")
    GEZ = wapordl.bounding_boxes.L3_BBS.pop("GEZ")
    _ = wapor_map(region_GEZ, "L3-E-D", period, tmp_path)
    assert wapordl.bounding_boxes.L3_BBS.get("GEZ", None) == GEZ


def test_18(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_MUV = os.path.join(test_data_folder, "test_MUV_noCRS.geojson")
    _ = wapor_map(region_MUV, "L3-T-D", period, tmp_path)


def test_19(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_MULTIPLE = os.path.join(test_data_folder, "test_MULTIPLE.geojson")
    _ = wapor_map(region_MULTIPLE, "L3-T-D", period, tmp_path)


def test_20(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_3D = os.path.join(test_data_folder, "test_3D.geojson")
    _ = wapor_map(region_3D, "L1-T-D", period, tmp_path)
    _ = wapor_ts(region_3D, "L1-T-D", period, overview=2)


def test_21(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_not_4326 = os.path.join(test_data_folder, "test_MUV_UTM36N.geojson")
    _ = wapor_map(region_not_4326, "L1-T-D", period, tmp_path)


def test_22(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_noCRS = os.path.join(test_data_folder, "test_MUV_noCRS.geojson")
    _ = wapor_map(region_noCRS, "L1-T-D", period, tmp_path)


def test_23(tmp_path):
    period = ["2021-01-12", "2021-01-25"]
    region_shpfile = os.path.join(
        test_data_folder, "test_MUV_UTM36N_shp/test_MUV_UTM36N.shp"
    )
    _ = wapor_map(region_shpfile, "L1-T-D", period, tmp_path)

#####
# UNIT CONVERSION CHECKS
#####


def test_from_dekad():
    region = os.path.join(test_data_folder, "1237500.geojson")
    periodX = ["2021-01-01", "2021-01-31"]
    overview = 3

    # FROM DEKAD
    df_dekad_ref = wapor_ts(region, "L2-AETI-D", periodX, overview=overview)
    df_dekad_day = wapor_ts(
        region, "L2-AETI-D", periodX, overview=overview, unit_conversion="day"
    )
    assert df_dekad_day.attrs["units"] == "mm/day"
    assert np.all(df_dekad_ref["mean"] == df_dekad_day["mean"])
    df_dekad_dekad = wapor_ts(
        region, "L2-AETI-D", periodX, overview=overview, unit_conversion="dekad"
    )
    assert np.all(
        np.isclose(
            df_dekad_ref["mean"] * df_dekad_ref.number_of_days.dt.days,
            df_dekad_dekad["mean"],
            atol=0,
            rtol=1e-3,
        )
    )
    assert df_dekad_dekad.attrs["units"] == "mm/dekad"
    df_dekad_month = wapor_ts(
        region, "L2-AETI-D", periodX, overview=overview, unit_conversion="month"
    )
    assert np.all(
        np.isclose(df_dekad_ref["mean"] * 31, df_dekad_month["mean"], atol=0, rtol=1e-3)
    )
    assert df_dekad_month.attrs["units"] == "mm/month"
    df_dekad_year = wapor_ts(
        region, "L2-AETI-D", periodX, overview=overview, unit_conversion="year"
    )
    assert np.all(
        np.isclose(df_dekad_ref["mean"] * 365, df_dekad_year["mean"], atol=0, rtol=1e-3)
    )
    assert df_dekad_year.attrs["units"] == "mm/year"


def test_from_month():
    region = os.path.join(test_data_folder, "1237500.geojson")
    periodX = ["2021-01-01", "2021-01-31"]
    overview = 3

    df_month_ref = wapor_ts(region, "L2-AETI-M", periodX, overview=overview)
    df_month_day = wapor_ts(
        region, "L2-AETI-M", periodX, overview=overview, unit_conversion="day"
    )
    assert df_month_day.attrs["units"] == "mm/day"
    assert np.all(
        np.isclose(df_month_ref["mean"] / 31, df_month_day["mean"], atol=0, rtol=1e-2)
    )
    df_month_dekad = wapor_ts(
        region, "L2-AETI-M", periodX, overview=overview, unit_conversion="dekad"
    )
    assert np.all(
        np.isclose(df_month_ref["mean"] / 3, df_month_dekad["mean"], atol=0, rtol=1e-2)
    )
    assert df_month_dekad.attrs["units"] == "mm/dekad"
    df_month_month = wapor_ts(
        region, "L2-AETI-M", periodX, overview=overview, unit_conversion="month"
    )
    assert np.all(df_month_ref["mean"] == df_month_month["mean"])
    assert df_month_month.attrs["units"] == "mm/month"
    df_month_year = wapor_ts(
        region, "L2-AETI-M", periodX, overview=overview, unit_conversion="year"
    )
    assert np.all(
        np.isclose(df_month_ref["mean"] * 12, df_month_year["mean"], atol=0, rtol=1e-3)
    )
    assert df_month_year.attrs["units"] == "mm/year"


def test_from_year():
    region = os.path.join(test_data_folder, "1237500.geojson")
    periodX = ["2021-01-01", "2021-01-31"]
    overview = 3

    df_year_ref = wapor_ts(region, "L2-AETI-A", periodX, overview=overview)
    df_year_day = wapor_ts(
        region, "L2-AETI-A", periodX, overview=overview, unit_conversion="day"
    )
    assert df_year_day.attrs["units"] == "mm/day"
    assert np.all(
        np.isclose(df_year_ref["mean"] / 365, df_year_day["mean"], atol=0, rtol=1e-2)
    )
    df_year_dekad = wapor_ts(
        region, "L2-AETI-A", periodX, overview=overview, unit_conversion="dekad"
    )
    assert np.all(
        np.isclose(df_year_ref["mean"] / 36, df_year_dekad["mean"], atol=0, rtol=1e-2)
    )
    assert df_year_dekad.attrs["units"] == "mm/dekad"
    df_year_month = wapor_ts(
        region, "L2-AETI-A", periodX, overview=overview, unit_conversion="month"
    )
    assert np.all(
        np.isclose(df_year_ref["mean"] / 12, df_year_month["mean"], atol=0, rtol=1e-3)
    )
    assert df_year_month.attrs["units"] == "mm/month"
    df_year_year = wapor_ts(
        region, "L2-AETI-A", periodX, overview=overview, unit_conversion="year"
    )
    assert np.all(df_year_ref["mean"] == df_year_year["mean"])
    assert df_year_year.attrs["units"] == "mm/year"


def test_unit_conversion_1():
    region = os.path.join(test_data_folder, "1237500.geojson")
    periodX = ["2021-01-01", "2021-01-31"]
    overview = 3

    _ = wapor_ts(region, "L1-T-D", periodX, overview=overview, unit_conversion="dekad")
    df_npp_dekad_per_dekad = wapor_ts(
        region, "L1-NPP-D", periodX, overview=overview, unit_conversion="dekad"
    )
    assert np.all(df_npp_dekad_per_dekad["minimum"]) >= 0
    df_rsm_dekad_per_dekad = wapor_ts(
        region, "L1-RSM-D", periodX, overview=overview, unit_conversion="dekad"
    )
    assert np.all(df_rsm_dekad_per_dekad["minimum"]) >= 0.0
    assert np.all(df_rsm_dekad_per_dekad["maximum"]) <= 1.0


def test_unit_conversion_2(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    periodX = ["2021-01-01", "2021-01-31"]

    fp19 = wapor_map(region, "L2-AETI-D", periodX, tmp_path, unit_conversion="dekad")
    ds = gdal.Open(fp19)
    assert ds.RasterCount == 3
    band = ds.GetRasterBand(1)
    ndv = band.GetNoDataValue()
    scale = band.GetScale()
    assert not isinstance(scale, type(None))
    array = band.ReadAsArray() * scale
    array[array == ndv * scale] = np.nan
    mean = np.nanmean(array)
    md = band.GetMetadata()
    assert np.all(
        [
            md[k] == v
            for k, v in {
                "end_date": "2021-01-10",
                "long_name": "Actual EvapoTranspiration and Interception",
                "number_of_days": "10",
                "original_units": "mm/day",
                "overview": "NONE",
                "start_date": "2021-01-01",
                "temporal_resolution": "Dekad",
                "units": "mm/dekad",
                "units_conversion_factor": "10",
            }.items()
        ]
    )
    assert mean > 0.0
    assert mean < 25.0
    proj = osr.SpatialReference(wkt=ds.GetProjection())
    assert proj.GetAttrValue("AUTHORITY", 1) == "4326"
    ds = ds.FlushCache()


def test_unit_conversion_3(tmp_path):
    region = os.path.join(test_data_folder, "1237500.geojson")
    period = ["2021-01-12", "2021-01-25"]

    fp20 = wapor_map(
        region, "L2-AETI-D", period, tmp_path, extension=".nc", unit_conversion="dekad"
    )
    info = gdal.Info(fp20, format="json")
    assert len(info["metadata"]["SUBDATASETS"]) == 4
    ds = gdal.Open(info["metadata"]["SUBDATASETS"]["SUBDATASET_1_NAME"])
    band = ds.GetRasterBand(1)
    scale = band.GetScale()
    assert not isinstance(scale, type(None))
    array = band.ReadAsArray() * scale
    ndv = band.GetNoDataValue()
    array[array == ndv * scale] = np.nan
    mean = np.nanmean(array)
    assert mean > 0.0
    assert mean < 15.0
    proj = osr.SpatialReference(wkt=ds.GetProjection())
    assert proj.GetAttrValue("AUTHORITY", 1) == "4326"
    ds = ds.FlushCache()


if __name__ == "__main__":
    tmp_path = r"/Users/hmcoerver/Local/test"
